<?php

/**
 * @file
 * Contains \LakeDrops\Drupal8Theme\SASS\Handler.
 */

namespace LakeDrops\Drupal8Theme\SASS;

use Composer\Package\PackageInterface;
use Composer\Package\Package;
use Composer\Composer;
use Composer\IO\IOInterface;
use Composer\Script\Event as ScriptEvent;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Finder\Finder;
use MatthiasMullie\PathConverter\Converter;

class Handler
{

  /**
   * @var \Composer\Composer
   */
  protected $composer;

  /**
   * @var \Composer\IO\IOInterface
   */
  protected $io;

  /**
   * @var \Twig_Loader_Array
   */
  protected $twig_loader;

  /**
   * @var \Twig_Environment
   */
  protected $twig;

  /**
   * @var \Composer\Installer\InstallationManager
   */
  protected $installationManager;

  /**
   * @var string
   */
  protected $installationPath;

  /**
   * @var array
   */
  protected $options;

  /**
   * @var bool
   */
  protected $update;

  /**
   * @var bool
   */
  protected $reset;

  /**
   * @var bool
   */
  protected $overwrite;

  /**
   * Handler constructor.
   *
   * @param Composer $composer
   * @param IOInterface $io
   */
  public function __construct(Composer $composer, IOInterface $io, $update = FALSE, $reset = FALSE, $overwrite = FALSE)
  {
    $this->composer = $composer;
    $this->io = $io;
    $this->twig_loader = new \Twig_Loader_Array([]);
    $this->twig = new \Twig_Environment($this->twig_loader);
    $this->update = $update;
    $this->reset = $reset;
    $this->overwrite = $overwrite;
  }

  public function setUpdate($flag) {
    $this->update = $flag;
  }

  /**
   * @return \Composer\Installer\InstallationManager
   */
  protected function getInstallationManager()
  {
    if (!isset($this->installationManager)) {
      $this->installationManager = $this->composer->getInstallationManager();
    }
    return $this->installationManager;
  }

  protected function buildInstallationPath($name) {
    $package = new Package($name, '1.0.0', '1.0.0');
    $package->setType('drupal-custom-theme');
    return $this->getInstallationManager()
      ->getInstallPath($package);
  }

  /**
   * @return string
   */
  protected function getInstallationPath()
  {
    if (!isset($this->installationPath)) {
      $this->installationPath = $this->buildInstallationPath($this->getOption('project_name'));
    }
    return $this->installationPath;
  }

  /**
   * Retrieve a package from the current composer process.
   *
   * @param string $name
   *   Name of the package to get from the current composer installation.
   *
   * @return PackageInterface
   */
  protected function getPackage($name)
  {
    return $this->composer->getRepositoryManager()
      ->getLocalRepository()
      ->findPackage($name, '*');
  }

  /**
   * Retrieve excludes from optional "extra" configuration.
   *
   * @return array
   */
  protected function getOptions()
  {
    if (!isset($this->options)) {
      $extra = $this->composer->getPackage()
          ->getExtra() + ['theme-d8-sass' => []];
      $this->options = array_replace_recursive([
        'project_name' => 'mytheme',
        'base_theme' => [
          'name' => 'bartik',
          'import' => [],
        ],
        'bower_assets' => [],
      ], $extra['theme-d8-sass']);

      $fs = new Filesystem();

      $dependencies = [];
      $jsFiles = [];
      $to = $this->buildInstallationPath($this->options['project_name']);
      foreach ($this->options['bower_assets'] as $key => $asset) {
        $package = $this->getPackage('bower-asset/' . $key);
        $version = $package ? $package->getVersion() : (isset($asset['version']) ? $asset['version'] : 'latest');
        if ($package) {
          // Installed globally as Composer package
          $from = $this->getInstallationManager()->getInstallPath($package);
          if (!$fs->exists($from)) {
            // Global package is missing
            unset($this->options['bower_assets'][$key]);
            continue;
          }
        }
        else {
          // Will be installed locally via Bower
          $from = implode(DIRECTORY_SEPARATOR, [$to, 'bower_components', $key]);
          $dependencies[] = '"' . $key . '": "' . $version . '"';
        }
        $converter = new Converter($from, $to);
        foreach (['fonts', 'sass', 'js'] as $type) {
          if (isset($asset[$type])) {
            $this->options['bower_assets'][$key][$type]['relative'] = $converter->convert($asset[$type]['src']);
            if ($type == 'js') {
              $jsFiles[$asset['name']] = [];
              $finder = new Finder();
              $finder->ignoreDotFiles(FALSE);
              $finder->files()->in($from . DIRECTORY_SEPARATOR . $asset[$type]['src']);
              foreach ($finder as $file) {
                $jsFiles[$asset['name']][] = $file->getRelativePathname();
              }
            }
            if ($this->reset && in_array($type, ['fonts', 'js'])) {
              $path = implode(DIRECTORY_SEPARATOR, [$to, $type, $asset['name']]);
              if ($fs->exists($path)) {
                $fs->remove($path);
              }
            }
          }
        }
      }
      $this->options['bower_assets']['dependencies'] = implode(",\n    ", $dependencies);
      $this->options['bower_assets']['jsfiles'] = $jsFiles;
    }
    return $this->options;
  }

  /**
   * @param string $key
   * @return string
   */
  protected function getOption($key)
  {
    $options = $this->getOptions();
    return $options[$key];
  }

  protected function getCustomFiles()
  {
    return [
      'config/install/{{ project_name }}.settings.yml',
      'config/schema/{{ project_name }}.schema.yml',
      'js/{{ project_name }}.js',
      'sass/_default-variables.scss',
      'sass/style.scss',
      '.gitignore',
      'favicon.ico',
      'logo.svg',
      'screenshot.png',
      '{{ project_name }}.info.yml',
      '{{ project_name }}.libraries.yml',
      '{{ project_name }}.theme',
    ];
  }

  protected function getIgnoredFiles()
  {
    return array_merge((empty($this->getOption('bower_assets')['dependencies']) ? ['bower.json'] : []), [
      'templates/README.md',
      'README.md',
      '{{ project_name }}.starterkit.yml',
      '.gitignore',
    ]);
  }

  protected function getBinaryFiles()
  {
    return [
      'favicon.ico',
      'screenshot.png',
    ];
  }

  protected function tweaks($string) {

    $string = str_replace('THEMENAME', '{{ project_name }}', $string);

    $baseTheme = $this->getOption('base_theme');
    if (isset($baseTheme['sasspath'])) {
      if (strpos($string, $baseTheme['sasspath']) === 0) {
        $string = 'sass' . substr($string, strlen($baseTheme['sasspath']));
      }
    }

    return $string;
  }

  protected function twig($string)
  {
    if (empty($string)) {
      return '';
    }

    $this->twig_loader->setTemplate('fixed', $this->tweaks($string));
    return str_replace('&quot;', '"', $this->twig->render('fixed', $this->getOptions()));
  }

  /**
   * Wrapper for npm command.
   *
   * @param $command
   *   NPM command name, arguments and/or options.
   * @throws \Exception
   */
  protected function npm($command)
  {
    passthru(escapeshellcmd('npm ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('NPM returned a non-zero exit code');
    }
  }

  /**
   * Wrapper for bower command.
   *
   * @param $command
   *   Bower command name, arguments and/or options.
   * @throws \Exception
   */
  protected function bower($command)
  {
    passthru(escapeshellcmd('bower --allow-root ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('Bower returned a non-zero exit code');
    }
  }

  /**
   * Wrapper for gulp command.
   *
   * @param $command
   *   Gulp command name, arguments and/or options.
   * @throws \Exception
   */
  protected function gulp($command)
  {
    passthru(escapeshellcmd('node_modules/.bin/gulp ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('Gulp returned a non-zero exit code');
    }
  }

  /**
   * @param FileSystem $fs
   * @param string $source
   * @param string $destination
   * @return array
   */
  protected function copyFiles($fs, $source, $destination) {
    $files = [];

    // Make sure all directories do exist
    $finder = new Finder();
    $finder->ignoreDotFiles(FALSE);
    $finder->directories()->in($source);
    foreach ($finder as $file) {
      $path = $destination . DIRECTORY_SEPARATOR . $this->twig($file->getRelativePathname());
      if (!$fs->exists($path)) {
        $fs->mkdir($path, 0775);
      }
    }

    $finder = new Finder();
    $finder->ignoreDotFiles(FALSE);
    $finder->files()->in($source);
    foreach ($finder as $file) {

      // Do not copy .gitkeep files
      if ($file->getFilename() == '.gitkeep') {
        continue;
      }

      // Determine relative filename of destination
      $relFilename = $file->getRelativePathname();
      if ($file->getExtension() == 'twig') {
        $relFilename = substr($relFilename, 0, strrpos($relFilename, '.twig'));
      }
      $tweakedRelFilename = $this->tweaks($relFilename);
      $filename = $destination . DIRECTORY_SEPARATOR . $this->twig($relFilename);

      // Do not copy ignored files
      if (in_array($tweakedRelFilename, $this->getIgnoredFiles())) {
        continue;
      }
      // Do not update custom files
      if (!$this->overwrite && $fs->exists($filename) && in_array($tweakedRelFilename, $this->getCustomFiles())) {
        continue;
      }

      // Now create destination
      if (in_array($tweakedRelFilename, $this->getBinaryFiles())) {
        $fs->copy($file->getRealPath(), $filename);
      } else {
        file_put_contents($filename, $this->twig(file_get_contents($file->getRealPath())));
      }
      $files[] = $tweakedRelFilename;
    }

    return $files;
  }

  /**
   * @param ScriptEvent $event
   */
  public function updateTheme($event)
  {

    $fs = new Filesystem();
    $destination = $this->getInstallationPath();

    $pluginRoot = $this->getInstallationManager()
      ->getInstallPath($this->getPackage('lakedrops/theme-d8-sass'));

    // Copy all files
    $this->copyFiles($fs, $pluginRoot . '/templates', $destination);

    // Handle optional base theme
    $baseTheme = $this->getOption('base_theme');
    if (isset($baseTheme['package'])) {
      $baseThemePackage = $this->getPackage($baseTheme['package']);
      if ($baseThemePackage && isset($baseTheme['starterkit'])) {
        $starterkitRoot = $this->getInstallationManager()->getInstallPath($baseThemePackage) . DIRECTORY_SEPARATOR . $baseTheme['starterkit'];
        $this->options['base_theme_files'] = $this->copyFiles($fs, $starterkitRoot , $destination);
      }
    }

    // Remove bower stuff if that's available globally
    if (empty($this->getOption('bower_assets')['dependencies'])) {
      if ($fs->exists($destination . DIRECTORY_SEPARATOR . 'bower.json')) {
        $fs->remove($destination . DIRECTORY_SEPARATOR . 'bower.json');
      }
      if ($fs->exists($destination . DIRECTORY_SEPARATOR . 'bower_components')) {
        $fs->remove($destination . DIRECTORY_SEPARATOR . 'bower_components');
      }
    }

    // Finally write .gitignore
    file_put_contents($destination . '/.gitignore', $this->twig(file_get_contents($pluginRoot . '/templates/.gitignore.twig')));

    // Prepare post processing
    $currentDir = getcwd();
    chdir($destination);

    // Handle NPM
    if ($this->update) {
      $this->npm('update');
    }
    else {
      $this->npm('install');
    }

    // Handle Bower if not installed globally
    if (!empty($this->getOption('bower_assets')['dependencies']) && $fs->exists($destination . DIRECTORY_SEPARATOR . 'bower.json')) {
      if ($this->update) {
        $this->bower('update');
      } else {
        $this->bower('install');
      }
    }

    // Run Gulp
    $args = ['default'];
    if (!$event->isDevMode()) {
      $args[] = '--env production';
    }
    $this->gulp(implode(' ', $args));

    // Restore current directory
    chdir($currentDir);
  }

}
