<?php

/**
 * @file
 * Contains \LakeDrops\Drupal8Theme\SASS\Tests\PluginTest.
 */

namespace LakeDrops\Drupal8Theme\SASS\Tests;

use Composer\Util\Filesystem;
use Symfony\Component\Yaml\Dumper;
use Symfony\Component\Yaml\Parser;

/**
 * Tests composer plugin functionality.
 */
class PluginTest extends \PHPUnit_Framework_TestCase {

  static $name = 'theme-d8-sass';

  /**
   * @var \Composer\Util\Filesystem
   */
  protected $fs;

  /**
   * @var string
   */
  protected $tmpDir;

  /**
   * @var string
   */
  protected $rootDir;

  /**
   * @var string
   */
  protected $tmpReleaseTag;

  /**
   * SetUp test
   */
  public function setUp() {
    $this->rootDir = realpath(realpath(__DIR__ . '/..'));

    // Prepare temp directory.
    $this->fs = new Filesystem();
    $this->tmpDir = realpath(sys_get_temp_dir()) . DIRECTORY_SEPARATOR . self::$name;
    $this->ensureDirectoryExistsAndClear($this->tmpDir);

    $this->writeTestReleaseTag();
    $this->writeComposerJSON();

    chdir($this->tmpDir);
  }

  /**
   * tearDown
   *
   * @return void
   */
  public function tearDown() {
    #$this->fs->removeDirectory($this->tmpDir);
    $this->git(sprintf('tag -d "%s"', $this->tmpReleaseTag));
  }

  /**
   * Tests a simple composer install without core, but adding core later.
   */
  public function testComposerInstallAndUpdate() {
    $infofile = $this->tmpDir . '/web/themes/custom/test_theme/test_theme.info.yml';
    $this->assertFileNotExists($infofile, 'Info file should not exist.');
    $this->composer('install');
    $this->assertFileExists($infofile, 'Info file should be automatically installed.');
    $this->fs->remove($infofile);
    $this->assertFileNotExists($infofile, 'Info file should not exist.');
    $this->composer('update');
    $this->assertFileExists($infofile, 'Info file should be automatically installed.');
    $this->fs->remove($infofile);
    $this->assertFileNotExists($infofile, 'Info file should not exist.');
    $this->composer(self::$name);
    $this->assertFileExists($infofile, 'Info file should be installed by "' . self::$name . '" command.');

    $yaml = new Parser();
    $dumper = new Dumper();

    $info = $yaml->parse(file_get_contents($infofile));
    $this->assertArrayNotHasKey('test', $info, 'Info file should not contain a test key.');

    $test_string = str_shuffle('abcdefghijklmnopqrstuvwxyz');
    $info['test'] = $test_string;
    file_put_contents($infofile, $dumper->dump($info, 2));

    $info = $yaml->parse(file_get_contents($infofile));
    $this->assertArrayHasKey('test', $info, 'Info file should contain a test key.');
    $this->assertEquals($test_string, $info['test']);
    $this->composer(self::$name);
    $info = $yaml->parse(file_get_contents($infofile));
    $this->assertArrayHasKey('test', $info, 'Info file should contain a test key.');
    $this->assertEquals($test_string, $info['test']);
  }

  /**
   * Writes the default composer json to the temp direcoty.
   */
  protected function writeComposerJSON() {
    $json = json_encode($this->composerJSONDefaults(), JSON_PRETTY_PRINT);
    file_put_contents($this->tmpDir . '/composer.json', $json);
  }

  /**
   * Writes a tag for the current commit, so we can reference it directly in the
   * composer.json.
   */
  protected function writeTestReleaseTag() {
    $this->tmpReleaseTag = '999.0.' . time();
    $this->git(sprintf('tag -a "%s" -m "%s"', $this->tmpReleaseTag, 'Tag for testing this exact commit'));
  }

  /**
   * Provides the default composer.json data.
   *
   * @return array
   */
  protected function composerJSONDefaults() {
    return array(
      'repositories' => array(
        array(
          'type' => 'vcs',
          'url' => $this->rootDir,
        ),
        array(
          'type' => 'composer',
          'url' => 'https://packages.drupal.org/8',
        ),
        array(
          'type' => 'composer',
          'url' => 'https://asset-packagist.org',
        ),
      ),
      'require' => array(
        'lakedrops/' . self::$name => $this->tmpReleaseTag,
        'composer/installers' => '~1.2.0',
        'drupal/bootstrap' => '3.2.0',
        'bower-asset/bootstrap-sass' => '~3.3.7',
        'bower-asset/flat-ui-sass' => '~2.1.3',
      ),
      'scripts' => array(
        self::$name => 'LakeDrops\\Drupal8Theme\\SASS\\Plugin::update',
      ),
      'minimum-stability' => 'dev',
      'extra' => array(
        'installer-paths' => array(
          'web/themes/custom/{$name}' => array(
            'type:drupal-custom-theme',
          ),
          'web/themes/contrib/{$name}' => array(
            'type:drupal-theme',
          ),
        ),
        self::$name => array(
          'project_name' => 'test_theme',
          'base_theme' => array(
            'name' => 'bootstrap',
            'package' => 'drupal/bootstrap',
            'starterkit' => 'starterkits/sass',
            'sasspath'=> 'scss',
            'import'=> array('overrides', ),
          ),
          'bower_assets' => array(
            'bootstrap-sass' => array(
              'name' => 'bootstrap',
              'fonts' => array(
                'src' => 'assets/fonts/bootstrap',
              ),
              'sass' => array(
                'src' => 'assets/stylesheets',
                'import' => array('bootstrap', ),
              ),
            ),
          ),
        ),
      ),
    );
  }

  /**
   * Wrapper for the composer command.
   *
   * @param string $command
   *   Composer command name, arguments and/or options
   * @throws \Exception
   */
  protected function composer($command) {
    chdir($this->tmpDir);
    passthru(escapeshellcmd($this->rootDir . '/vendor/bin/composer ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('Composer returned a non-zero exit code');
    }
  }

  /**
   * Wrapper for git command in the root directory.
   *
   * @param $command
   *   Git command name, arguments and/or options.
   * @throws \Exception
   */
  protected function git($command) {
    chdir($this->rootDir);
    passthru(escapeshellcmd('git -c "user.email=phpunit@test.com" ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new \Exception('Git returned a non-zero exit code');
    }
  }

  /**
   * Makes sure the given directory exists and has no content.
   *
   * @param string $directory
   */
  protected function ensureDirectoryExistsAndClear($directory) {
    if (is_dir($directory)) {
      $this->fs->removeDirectory($directory);
    }
    mkdir($directory, 0777, TRUE);
  }

}
